#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h> /*для функции atoi()*/

/*Шаблон содержимого файла конфигурации формата XML*/
#define APPLICATION_XML_TEMPLATE "<?xml version=\"1.1\" encoding=\"UTF-8\" ?>\n<settings>\n\t<font>%s</font>\n\t<color>%s</color>\n\t<wrap>%i</wrap>\n\t<openlast>%i</openlast>\n\t<lastfile>%s</lastfile>\n</settings>"

/*Перечисление идентификаторов элементов файла конфигурации формата XML*/
enum {
	APPLICATION_XML_PARAMETER_FONT = 0,
	APPLICATION_XML_PARAMETER_COLOR,
	APPLICATION_XML_PARAMETER_WRAP,
	APPLICATION_XML_PARAMETER_OPENLAST,
	APPLICATION_XML_PARAMETER_LASTFILE,
	APPLICATION_XML_PARAMETER_UNKNOWN
};

/*Данные приложения*/
typedef struct {
	/*Объект приложения*/
	GtkApplication *app;
	/*Описание шрифта*/
	gchar *font;
	/*Описание цвета шрифта*/
	gchar *color;
	/*Идентификатор режима переноса строк*/
	gint wrap;
	/*Флаг открытия последнего открытого текстового файла при запуске приложения*/
	gboolean openlast;
	/*Путь к последнему открытому текстовому файлу*/
	gchar *lastfile;
	/*Идентификатор текущего параметра файла конфигурации формата XML*/
	guint xmlparameter;
	/*Путь к файлу конфигурации формата XML*/
	gchar *xmlfilename;
	/*Путь к открытому текстовому файлу*/
	gchar *textfilename;
	/*Виджет основного окна приложения*/
	GtkWidget *mainwin; 
	/*Виджет кнопки выбора шрифта*/
	GtkWidget *font_selector;
	/*Виджет кнопки выбора цвета шрифта*/
	GtkWidget *color_selector;
	/*Виджет раскрывающегося меню для выбора режима переноса строк*/
	GtkWidget *wrap_selector;
	/*Виджет переключателя для активации функции автоматического открытия файла*/
	GtkWidget *openfile_selector;
	/*Виджет поля для редактирования текста*/
	GtkWidget *file_contents_textview;
	/*Виджет строки состояния*/
	GtkWidget *statusbar;
	/*Идентификатор последнего сообщения виджета строки состояния*/
	guint statusmsgid;
	/*Идентификатор контекста виджета строки состояния*/
	guint statusconid;
} appdata;

/*Функция для форматированного вывода сообщения средствами виджета строки состояния*/
static void application_log_action(appdata *data, const gchar *action)
{
	GDateTime *datetime;
	guint h, m, s;
	gchar *statusmsg;
		
	g_assert(data != NULL);
	g_assert(action != NULL);
	
	datetime = g_date_time_new_now_local();
	if (datetime != NULL) {
		/*Получение значений часов, минут и секунд*/
		h = g_date_time_get_hour(datetime);
		m = g_date_time_get_minute(datetime);
		s = g_date_time_get_second(datetime);
		/*Форматирование строки*/
		if (data->textfilename != NULL) {
			statusmsg = g_strdup_printf("%u%u:%u%u:%u%u: Файл %s %s", h/10, h%10, m/10, m%10, s/10, s%10, data->textfilename, action);
		} else {
			statusmsg = g_strdup_printf("%u%u:%u%u:%u%u: Файл %s", h/10, h%10, m/10, m%10, s/10, s%10, action);
		}
		/*Удаление предыдущей строки*/
		if (data->statusmsgid != 0) {
			gtk_statusbar_remove(GTK_STATUSBAR(data->statusbar), data->statusconid, data->statusmsgid);
		}
		/*Вывод строки*/
		data->statusmsgid = gtk_statusbar_push(GTK_STATUSBAR(data->statusbar), data->statusconid, statusmsg);
		/*Освобождение ресурсов*/
		g_free(statusmsg);
		g_date_time_unref(datetime);
	} 
}

/*Функция для заменты стандартных сущностей на соответствующие им символы при получении строковых данных из файла формата XML*/
static gchar *application_xml_markup_unescape(const gchar *string, gsize length)
{
	gint srcptr, resptr, seqptr;
	gchar *res, *movedres;
	/*Набор стандартных сущностей и соответствующих им символов*/
	struct seqdesc {
		gchar *seq;
		gchar sym;
		gchar len;
	} seqrange[] = {{"&lt;", '<', 4}, {"&gt;", '>', 4}, {"&amp;", '&', 5}, {"&apos;", '\'', 6}, {"&quot;", '\"', 6}};
	    
	
	if ((string == NULL) || (length == 0)) return NULL;
	
	res = g_malloc0(length + 1);
	
	srcptr = 0;
	resptr = 0;
	
	while (srcptr < length) {
		/*Поиск символа '&'*/
		if (string[srcptr] == '&') {
			/*Установление использованной стандартной сущности*/
			for (seqptr=0; seqptr<sizeof(seqrange)/sizeof(struct seqdesc); seqptr++) {
				if (seqrange[seqptr].len <= (length - srcptr)) {
					if (g_ascii_strncasecmp(string + srcptr, seqrange[seqptr].seq, seqrange[seqptr].len) == 0) {
						/*Замена стандартной сущности на соответствующий ей символ*/
						res[resptr++] = seqrange[seqptr].sym;
						srcptr += seqrange[seqptr].len;
						continue;
					}
				}
			}
		}
		/*Простое копирование символа*/
		res[resptr++] = string[srcptr++];
	}
	
	/*Сокращение размера фрагмента памяти, использованного для хранения результирующей строки, если это возможно*/
	if (resptr < srcptr) {
		movedres = g_try_realloc(res, resptr);
		if (movedres != NULL) {
			res = movedres;
		}
	}
	
	return res;
}

/*Функция начала разбора элемента*/
static void application_xml_get_start_element(GMarkupParseContext *context, const gchar *element_name, const gchar **attribute_names, const gchar **attribute_values, gpointer user_data, GError **error)
{
	appdata *data;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Установка соответствующего идентификатора элемента для каждого из имен элементов*/
	if (g_str_equal(element_name, "settings")) {
		data->xmlparameter = APPLICATION_XML_PARAMETER_UNKNOWN;
	} else if (g_str_equal(element_name, "font")) {
		data->xmlparameter = APPLICATION_XML_PARAMETER_FONT;
	} else if (g_str_equal(element_name, "color")) {
		data->xmlparameter = APPLICATION_XML_PARAMETER_COLOR;
	} else if (g_str_equal(element_name, "wrap")) {
		data->xmlparameter = APPLICATION_XML_PARAMETER_WRAP;
	} else if (g_str_equal(element_name, "openlast")) {
		data->xmlparameter = APPLICATION_XML_PARAMETER_OPENLAST;
	} else if (g_str_equal(element_name, "lastfile")) {
		data->xmlparameter = APPLICATION_XML_PARAMETER_LASTFILE;
	} else {
		/*Неизвестный элемент*/
		data->xmlparameter = APPLICATION_XML_PARAMETER_UNKNOWN;
	}
}

/*Функция разбора текста*/
static void application_xml_get_value(GMarkupParseContext *context, const gchar *text, gsize text_len, gpointer user_data, GError **error)
{
	appdata *data;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Идентификатор неизвестного элемента*/
	if (data->xmlparameter == APPLICATION_XML_PARAMETER_UNKNOWN) return;
	
	switch (data->xmlparameter) {
		/*Получение строкового описания шрифта*/
		case APPLICATION_XML_PARAMETER_FONT:
			if (data->font != NULL) {
				g_free(data->font);
			}
			data->font = application_xml_markup_unescape(text, text_len);
			break;
		/*Получение строкового описания типа шрифта*/
		case APPLICATION_XML_PARAMETER_COLOR:
			if (data->color != NULL) {
				g_free(data->color);
			}
			data->color = application_xml_markup_unescape(text, text_len);
			break;
		/*Получение целочисленного идентификатора режима переноса строк*/
		case APPLICATION_XML_PARAMETER_WRAP:
			data->wrap = (gint)atoi(text);
			break;
		/*Получение логического значения флага открытия последнего открытого файла*/
		case APPLICATION_XML_PARAMETER_OPENLAST:
			data->openlast = (gboolean)atoi(text);
			break;
		/*Получение строки, являющейся путем к последнему открытому файлу*/
		case APPLICATION_XML_PARAMETER_LASTFILE:
			if (data->lastfile != NULL) {
				g_free(data->lastfile);
			}
			data->lastfile = application_xml_markup_unescape(text, text_len);
			break;
		default:
			break;
	}
	
}

/*Функция окончания разбора элемента*/
static void application_xml_get_end_element(GMarkupParseContext *context, const gchar *element_name, gpointer user_data, GError **error)
{
	appdata *data;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Сброс идентификатора элемента при закрытии корневого тэга */
	if (!g_str_equal(element_name, "settings")) {
		data->xmlparameter = APPLICATION_XML_PARAMETER_UNKNOWN;
	}
	
	return;
}

/*Функция для открытия и разбора файла конфигурации формата XML*/
static gboolean application_open_config_file(appdata *data)
{
	gchar *configdir;
	GMarkupParser parser;
	GMarkupParseContext *context;
	gchar *contents;
	gsize length;
	GError *error;
	
	g_assert(data != NULL);
	
	data->xmlfilename = NULL;
	
	/*Формирование пути к директории для хранения файла конфигурации формата XML*/
	configdir = g_build_path(G_DIR_SEPARATOR_S, g_get_tmp_dir(), "GMarkupParserContextTest", NULL);
	
	/*Проверка наличия директории и ее создание при необходимости*/
	if (!g_file_test((const gchar *)configdir, G_FILE_TEST_IS_DIR)) {
		if (g_mkdir_with_parents(configdir, 0755) == -1) {
			g_warning("Не удалось создать директорию для сохранения файла конфигурации\n");
			g_free(configdir);
			return FALSE;
		}
	}
	/*Формирование пути к файлу конфигурации формата XML*/
	data->xmlfilename = g_build_filename(configdir, "settings.xml", NULL);
	
	/*Проверка наличия файла формата XML и загрузка его содержимого*/
	if (g_file_test((const gchar *)data->xmlfilename, G_FILE_TEST_IS_REGULAR)) {
		error = NULL;
		if (g_file_get_contents((const gchar *)data->xmlfilename, &contents, &length, &error)) {
			/*Установка указателя на функцию начала разбора элемента*/
			parser.start_element = application_xml_get_start_element;
			/*Установка указателя на функцию окончания разбора элемента*/
			parser.end_element = application_xml_get_end_element;
			/*Установка указателя на функцию разбора текста*/ 
			parser.text = application_xml_get_value;
			/*Реализация функции пропуска текста не используется*/
			parser.passthrough = NULL;
			/*Реализация функции обработки сообщения об ошибке не используется*/
			parser.error = NULL;
			
			/*Создание объекта контекста разбора документа формата XML*/
			context = g_markup_parse_context_new(&parser, 0, data, NULL);
			/*Непосредственный разбор документа*/
			g_markup_parse_context_parse(context, contents, (gssize)length, &error);
			if (error != NULL) {
				g_warning("Ошибка при обработке файла формата XML: %s\n", error->message);
				g_error_free(error);
			}
			/*Уничтожение объекта контекста разбора документа формата XML*/
			g_markup_parse_context_free(context);
			/*Освобождение памяти, использованной для хранения содержимого файла*/
			g_free(contents);
		} else if (error != NULL) {
			g_warning("Ошибка при чтении файла формата XML: %s\n", error->message);
			g_error_free(error);
		}
	}
	/*Освобождение памяти, использованной для хранения пути к директории с файлом конфигурации формата XML*/
	g_free(configdir);
	
	return TRUE;
}

/*Функция для закрытия файла конфигурации формата XML и записи данных конфигурации на диск*/
static gboolean application_close_config_file(appdata *data)
{
	gchar *contents;
	GError *error;
	
	g_assert(data != NULL);
	
	if (data->xmlfilename == NULL) return FALSE;
	
	/*Формирование строкового представления содержимого файла формата XML*/
	contents = g_markup_printf_escaped(APPLICATION_XML_TEMPLATE, 
										data->font != NULL ? data->font : "",
										data->color != NULL ? data->color : "",
										data->wrap,
										data->openlast,
										data->lastfile != NULL ? data->lastfile : "");
	
	/*Запись строкового представления содержимого файла формата XML в файл на диске*/
	error = NULL;
	if (!g_file_set_contents((const gchar *)data->xmlfilename, (const gchar *)contents, -1, &error)) {
		g_warning("Не удалось записать данные файла формата XML на диск: %s\n", error->message);
		g_error_free(error);
		g_free(contents);
		g_free(data->xmlfilename);
		return FALSE;
	}
	/*Освобождение ресурсов*/
	g_free(contents);
	g_free(data->xmlfilename);
	
	return TRUE;
}

/*Функция для вывода установленных значений в диалоге редактирования параметров конфигурации приложения*/
static void application_show_settings(appdata *data)
{
	GdkRGBA rgba;
		
	g_assert(data != NULL);
	
	/*Вывод описания шрифта*/
	if (data->font != NULL) {
		gtk_font_button_set_font_name(GTK_FONT_BUTTON(data->font_selector), (const gchar *)data->font);
	}
	/*Вывод описания цвета шрифта*/
	if (data->color != NULL) {
		if (gdk_rgba_parse(&rgba, (const gchar *)data->color)) {
			gtk_color_chooser_set_rgba(GTK_COLOR_CHOOSER(data->color_selector), (const GdkRGBA *)&rgba);
		}
	}
	/*Вывод идентификатора режима переноса строк*/
	if ((data->wrap <= 3) && (data->wrap >= 0)) {
		gtk_combo_box_set_active(GTK_COMBO_BOX(data->wrap_selector), data->wrap);
	} else {
		gtk_combo_box_set_active(GTK_COMBO_BOX(data->wrap_selector), 0);
	}
	/*Вывод флага автоматического открытия файла*/
	gtk_switch_set_active(GTK_SWITCH(data->openfile_selector), data->openlast);
}

/*Функция для применения значенний параметров конфигурации приложения*/
static void application_apply_settings(appdata *data, gboolean initialization)
{
	PangoFontDescription *fontdesc; 
	GdkRGBA rgba;
	gchar *contents;
	gsize length;
	GtkTextBuffer *buffer;
		
	g_assert(data != NULL);
	
	/*Применение описания шрифта*/
	if (data->font != NULL) {
		fontdesc = pango_font_description_from_string(data->font);
		if (fontdesc != NULL) {
			gtk_widget_override_font(data->file_contents_textview, fontdesc);
			pango_font_description_free(fontdesc);
		}
	}
	/*Применение описания цвета шрифта*/
	if (data->color != NULL) {
		if (gdk_rgba_parse(&rgba, (const gchar *)data->color)) {
			gtk_widget_override_color(data->file_contents_textview, GTK_STATE_FLAG_NORMAL, &rgba);
		}
	}
	/*Применение идентификатора переноса строк*/
	switch (data->wrap) {
		case 0:
			gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(data->file_contents_textview), GTK_WRAP_WORD);
			break;
		case 1:
			gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(data->file_contents_textview), GTK_WRAP_CHAR);
			break;
		case 2:
			gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(data->file_contents_textview), GTK_WRAP_WORD_CHAR);
			break;
		case 3:
			gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(data->file_contents_textview), GTK_WRAP_NONE);
			break;
		default:
			gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(data->file_contents_textview), GTK_WRAP_WORD);
			break;
	}
	/*Открытие файла при необходимости в случае запуска приложения*/
	if ((initialization) && (data->openlast)) {
		data->textfilename = g_strdup(data->lastfile);
		if (data->textfilename != NULL) {
			/*Загрузка и вывод содержимого файла*/
			if (g_file_get_contents((const gchar *)data->textfilename, &contents, &length, NULL)) {
				buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(data->file_contents_textview));
				gtk_text_buffer_set_text(buffer, (const gchar *)contents, length);
				g_free(contents);
				/*Вывод сообщения об открытии файла*/
				application_log_action(data, "открыт автоматически");
			}
		}
	}
}

/*Функция для сохранения значений параметров конфигурации приложения*/
static void application_save_settings(GtkDialog *dialog, gint response_id, gpointer user_data)
{
	appdata *data;
	GdkRGBA rgba;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Сохранение описания шрифта*/
	if (data->font != NULL) {
		g_free(data->font);
	}
	data->font = g_strdup(gtk_font_button_get_font_name(GTK_FONT_BUTTON(data->font_selector)));
	/*Сохранение описания цвета шрифта*/
	if (data->color != NULL) {
		g_free(data->color);
	}
	gtk_color_chooser_get_rgba(GTK_COLOR_CHOOSER(data->color_selector), &rgba);
	data->color = gdk_rgba_to_string((const GdkRGBA *)&rgba);
	/*Сохранение идентификатора режима переноса строк*/
	data->wrap = gtk_combo_box_get_active(GTK_COMBO_BOX(data->wrap_selector));
	/*Сохранение флага открытия последнего открытого файла при запуске приложения*/
	data->openlast = gtk_switch_get_active(GTK_SWITCH(data->openfile_selector));	
	/*Применение пзначений параметров конфигурации приложения*/
	application_apply_settings(data, FALSE);
	/*Уничтожение окна диалога изменения параметров конфгурации приложения*/
	gtk_widget_destroy(GTK_WIDGET(dialog));
}

/*Функция для сохранения пути к последнему открытому файлу*/
static void application_save_opened_file(appdata *data)
{
	g_assert(data != NULL);
	g_assert(data->textfilename != NULL);
	
	if (data->lastfile != NULL) {
		g_free(data->lastfile);
	}
	data->lastfile = g_strdup(data->textfilename);
}

/*Функция для завершения работы приложения*/
static void application_quit(GtkToolButton *toolbutton, gpointer user_data)
{
	appdata *data;
	
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	g_application_quit(G_APPLICATION(data->app));	
}

/*Функция для создания диалога для редактирования параметров конфигурации приложения*/
static void application_preferences(GtkToolButton *toolbutton, gpointer user_data)
{
	appdata *data;
	GtkWidget *dialog, *content;
	GtkGrid *grid;
	GtkWidget *label;
	
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Создание диалога для редактирования параметров конфгурации приложения*/
	dialog = gtk_dialog_new_with_buttons("Параметры",
										GTK_WINDOW(data->mainwin),
										GTK_DIALOG_MODAL|GTK_DIALOG_DESTROY_WITH_PARENT,
										"Закрыть",
										GTK_RESPONSE_NONE,
										NULL);
	
	content = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
	/*Создание сетки с виджетами для редактирования параметров конфигурации приложения*/
	grid = (GtkGrid *)gtk_grid_new();
	gtk_grid_set_row_spacing(grid, 5);
	gtk_grid_set_column_spacing(grid, 5);
	gtk_container_set_border_width(GTK_CONTAINER(grid), 10);
	
	label = gtk_label_new("Шрифт редактора");
	gtk_widget_set_halign(label, GTK_ALIGN_START);
	gtk_grid_attach(grid, label, 0, 0, 1, 1);
	
	data->font_selector = gtk_font_button_new();
	gtk_grid_attach(grid, data->font_selector, 1, 0, 1, 1);
	
	label = gtk_label_new("Цвет символов");
	gtk_widget_set_halign(label, GTK_ALIGN_START);
	gtk_grid_attach(grid, label, 0, 1, 1, 1);
	
	data->color_selector = gtk_color_button_new();
	gtk_grid_attach(grid, data->color_selector, 1, 1, 1, 1);
	
	label = gtk_label_new("Разрывы строк");
	gtk_widget_set_halign(label, GTK_ALIGN_START);
	gtk_grid_attach(grid, label, 0, 2, 1, 1);
	
	data->wrap_selector = gtk_combo_box_text_new();
	gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(data->wrap_selector), "По словам");
	gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(data->wrap_selector), "По символам");
	gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(data->wrap_selector), "По символам и словам");
	gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(data->wrap_selector), "Не используются");
	gtk_grid_attach(grid, data->wrap_selector, 1, 2, 1, 1);
	
	label = gtk_label_new("Открывать последний файл");
	gtk_widget_set_halign(label, GTK_ALIGN_START);
	gtk_grid_attach(grid, label, 0, 3, 1, 1);
	
	data->openfile_selector = gtk_switch_new();
	gtk_grid_attach(grid, data->openfile_selector, 1, 3, 1, 1);
	/*Показ загруженных параметров конфигурации приложения*/
	application_show_settings(data);
	/*Сигнал для сохранения настроек и уничтожения диалога*/
	g_signal_connect(dialog, "response", G_CALLBACK(application_save_settings), data);
	/*Добавление сетки с виджетами в окно приложения и показ диалога*/
	gtk_container_add(GTK_CONTAINER(content), GTK_WIDGET(grid));
	gtk_widget_show_all(dialog);
}

/*Функция для сохранения текстового файла*/
static void application_save_file(GtkToolButton *toolbutton, gpointer user_data)
{
	appdata *data;
	GtkWidget *dialog;
	GtkTextBuffer *buffer;
	GtkTextIter start, end;
	gchar *contents;
	GError *error;
	
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	if (data->textfilename == NULL) {
		/*Создание диалога сохранения файла*/
		dialog = gtk_file_chooser_dialog_new("Сохранить файл",
											GTK_WINDOW(data->mainwin),
											GTK_FILE_CHOOSER_ACTION_SAVE,
											"Отмена", GTK_RESPONSE_CANCEL,
											"Сохранить", GTK_RESPONSE_ACCEPT,
											NULL);
		
		gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(dialog), TRUE);
		gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(dialog), "file.txt");
		/*Получение пути к выбранному текстовому файлу*/
		if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
			data->textfilename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
		}
		/*Уничтожение диалога*/
		gtk_widget_destroy(dialog);
	}
	
	if (data->textfilename != NULL) {
		/*Получение содержимого виджета поля для редактирования текста*/
		buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(data->file_contents_textview));
		gtk_text_buffer_get_bounds(buffer, &start, &end);
		contents = gtk_text_buffer_get_text(buffer, (const GtkTextIter *)&start, (const GtkTextIter *)&end, FALSE);
		/*Запись полученных данных в файл на диск*/
		error = NULL;
		if (g_file_set_contents((const gchar *)data->textfilename, contents, -1, &error)) {
			/*Сохранение пути к текстовому файлу*/
			application_save_opened_file(data);
			/*Вывод сообщения о сохранении текстового файла*/
			application_log_action(data, "сохранен");
		} else if (error != NULL) {
			g_warning("Не удалось записать содержимое редактируемого файла на диск: %s\n", error->message);
			g_error_free(error);
		}
		/*Освобождение использованной памяти*/
		g_free(contents);
	}
}

/*Функция для открытия текстового файла*/
static void application_open_file(GtkToolButton *toolbutton, gpointer user_data)
{
	appdata *data;
	GtkWidget *dialog;
	GtkTextBuffer *buffer;
	GtkTextIter start, end;
	gchar *contents;
	gsize length;
	GError *error;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Создание диалога открытия файла*/
	dialog = gtk_file_chooser_dialog_new("Открыть файл",
										GTK_WINDOW(data->mainwin),
										GTK_FILE_CHOOSER_ACTION_SAVE,
										"Отмена", GTK_RESPONSE_CANCEL,
										"Открыть", GTK_RESPONSE_ACCEPT,
										NULL);
	
	if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
		if (data->textfilename != NULL) {
			g_free(data->textfilename);
		}
		/*Получение пути к выбранному текстовому файлу*/
		data->textfilename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
		/*Удаление содержимого поля для редактирования текста*/
		buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(data->file_contents_textview));
		gtk_text_buffer_get_bounds(buffer, &start, &end);
		gtk_text_buffer_delete(buffer, &start, &end);
		/*Загрузка данных из файла с диска и размещение их в виджете поля для редактирования текста*/
		error = NULL;
		if (g_file_get_contents((const gchar *)data->textfilename, &contents, &length, &error)) {
			gtk_text_buffer_set_text(buffer, (const gchar *)contents, length);
			/*Сохранение пути к текстовому файлу*/
			application_save_opened_file(data);
			g_free(contents);
			/*Вывод сообщения об открытии текстового файла*/
			application_log_action(data, "открыт");
		} else if (error != NULL) {
			g_warning("Не удалось прочитать содержимое файла с диска: %s\n", error->message);
			g_error_free(error);
		}
	}
	/*Уничтожение диалога открытия файла*/
	gtk_widget_destroy(dialog);
}

/*Функция для создания нового текстового файла*/
static void application_new_file(GtkToolButton *toolbutton, gpointer user_data)
{
	appdata *data;
	GtkTextBuffer *buffer;
	GtkTextIter start, end;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Удаление устаревшего пути к текстовому файлу*/
	if (data->textfilename != NULL) {
		g_free(data->textfilename);
		data->textfilename = NULL;
	}
	/*Удаление содержимого виджета поля для редактирования текста*/
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(data->file_contents_textview));
	gtk_text_buffer_get_bounds(buffer, &start, &end);
	gtk_text_buffer_delete(buffer, &start, &end);
	/*Вывод сообщения о создании текстового файла*/
	application_log_action(data, "создан");
}

/*Функция для создания главного окна приложения*/
static void application_activate(GtkApplication *application, gpointer user_data)
{
	appdata *data;
	GtkWidget *vbox;
	GtkWidget *toolbar;
	GtkToolItem *newtb, *opentb, *savetb, *septb1, *preftb, *septb2, *quittb;
	GtkWidget *scrollwin;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	application_open_config_file(data);
	/*Создание главного окна*/
	data->mainwin = gtk_application_window_new(application);
	gtk_window_set_title(GTK_WINDOW(data->mainwin), "Пример использования класса GMarkupParserContext");
	gtk_widget_set_size_request(data->mainwin, 600, 400);
	/*Создание контейнера для хранения элементов главного окна*/
	vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
	/*Создание панели инструментов и кнопок этой панели*/
	toolbar = gtk_toolbar_new();
	gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_BOTH_HORIZ);
	
	newtb = gtk_tool_button_new(NULL, "Создать");
	gtk_tool_button_set_icon_name(GTK_TOOL_BUTTON(newtb), "document-new");
	gtk_tool_item_set_is_important(GTK_TOOL_ITEM(newtb), TRUE);
	g_signal_connect(newtb, "clicked", G_CALLBACK(application_new_file), data);
	gtk_toolbar_insert(GTK_TOOLBAR(toolbar), GTK_TOOL_ITEM(newtb), -1);
		
	opentb = gtk_tool_button_new(NULL, "Открыть");
	gtk_tool_button_set_icon_name(GTK_TOOL_BUTTON(opentb), "document-open");
	gtk_tool_item_set_is_important(GTK_TOOL_ITEM(opentb), TRUE);
	g_signal_connect(opentb, "clicked", G_CALLBACK(application_open_file), data);
	gtk_toolbar_insert(GTK_TOOLBAR(toolbar), GTK_TOOL_ITEM(opentb), -1);
	
	savetb = gtk_tool_button_new(NULL, "Сохранить");
	gtk_tool_button_set_icon_name (GTK_TOOL_BUTTON (savetb), "document-save");
	gtk_tool_item_set_is_important(GTK_TOOL_ITEM(savetb), TRUE);
	g_signal_connect(savetb, "clicked", G_CALLBACK(application_save_file), data);
	gtk_toolbar_insert(GTK_TOOLBAR(toolbar), GTK_TOOL_ITEM(savetb), -1);
	
	septb1 = gtk_separator_tool_item_new();
	gtk_toolbar_insert(GTK_TOOLBAR(toolbar), septb1, -1);
	
	preftb = gtk_tool_button_new(NULL, "Параметры");
	gtk_tool_button_set_icon_name(GTK_TOOL_BUTTON(preftb), "preferences-system");
	gtk_tool_item_set_is_important(GTK_TOOL_ITEM(preftb), TRUE);
	g_signal_connect(preftb, "clicked", G_CALLBACK(application_preferences), data);
	gtk_toolbar_insert(GTK_TOOLBAR(toolbar), GTK_TOOL_ITEM(preftb), -1);
	
	septb2 = gtk_separator_tool_item_new();
	gtk_toolbar_insert(GTK_TOOLBAR(toolbar), septb2, -1);
	
	quittb = gtk_tool_button_new(NULL, "Выход");
	gtk_tool_button_set_icon_name (GTK_TOOL_BUTTON (quittb), "application-exit");
	gtk_tool_item_set_is_important(GTK_TOOL_ITEM(quittb), TRUE);
	g_signal_connect(quittb, "clicked", G_CALLBACK(application_quit), data);
	gtk_toolbar_insert(GTK_TOOLBAR(toolbar), GTK_TOOL_ITEM(quittb), -1);
	
	gtk_box_pack_start(GTK_BOX(vbox), toolbar, FALSE, TRUE, 0); 
	/*Создание окна для прокрутки содержимого виджета поля для редактирования текста*/
	scrollwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrollwin), GTK_SHADOW_IN);
	/*Создание виджета поля для редактирования текста*/
	data->file_contents_textview = gtk_text_view_new();
	gtk_container_add(GTK_CONTAINER(scrollwin), data->file_contents_textview);
	
	gtk_box_pack_start(GTK_BOX(vbox), scrollwin, TRUE, TRUE, 0);
	/*Создание виджета строки состояния*/
	data->statusbar = gtk_statusbar_new();
	data->statusmsgid = 0;
	data->statusconid = gtk_statusbar_get_context_id(GTK_STATUSBAR(data->statusbar), "GMarkupParserContextTest");
	gtk_box_pack_start(GTK_BOX(vbox), data->statusbar, FALSE, TRUE, 0);
		
	gtk_container_add(GTK_CONTAINER(data->mainwin), GTK_WIDGET(vbox));
	/*Применение загруженных значений параметров конфигурации приложения*/
	application_apply_settings(data, TRUE);
	/*Показ главного окна приложения*/
	gtk_application_add_window(application, GTK_WINDOW(data->mainwin));
	gtk_widget_show_all(data->mainwin);
}

/*Функция, вызываемая перед завершением работы приложения*/
static void application_shutdown(GtkApplication *application, gpointer user_data)
{
	appdata *data;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	/*Запись строкового представления документа формата XML в файл на диске*/
	application_close_config_file(data);
}

int main(int argc, char **argv)
{
	appdata *data;
	gint res;
	
	/*Создание структуры для хранения данных приложения*/
	data = g_new0(appdata, 1);
	/*Создание объекта приложения*/
	data->app = gtk_application_new("org.gtk.GMarkupParserContextTest", G_APPLICATION_FLAGS_NONE);
	/*Установка обработчиков основных сигналов объекта приложения*/
	g_signal_connect(data->app, "activate", G_CALLBACK(application_activate), data);
	g_signal_connect(data->app, "shutdown", G_CALLBACK(application_shutdown), data);
	/*Запуск основного цикла обработки событий приложения*/
	res = g_application_run(G_APPLICATION(data->app), 0, NULL);
	/*Освобождение памяти перед завершением работы приложения*/
	g_free(data);
	
	return res;
}
